disponible_environments=[
    "align"
    "align*"
    "alignat"
    "alignat*"
    "aligned"
    "alignedat"
    "array"
    "bmatrix"
    "Bmatrix"
    "bmatrix*"
    "Bmatrix*"
    "bsmallmatrix"
    "Bsmallmatrix"
    "bsmallmatrix*"
    "Bsmallmatrix*"
    "cases"
    "cases*"
    "CD"
    "crampedsubarray"
    "dcases"
    "dcases*"
    "drcases"
    "drcases*"
    "empheq"
    "eqnarray"
    "eqnarray*"
    "equation"
    "equation*"
    "flalign"
    "flalign*"
    "gather"
    "gather*"
    "gathered"
    "lgathered"
    "matrix"
    "matrix*"
    "multline"
    "multline*"
    "multlined"
    "numcases"
    "pmatrix"
    "pmatrix*"
    "prooftree"
    "psmallmatrix"
    "psmallmatrix*"
    "rcases"
    "rcases*"
    "rgathered"
    "smallmatrix"
    "smallmatrix*"
    "split"
    "spreadlines"
    "subarray"
    "subnumcases"
    "vmatrix"
    "Vmatrix"
    "vmatrix*"
    "Vmatrix*"
    "vsmallmatrix"
    "Vsmallmatrix"
    "vsmallmatrix*"
    "Vsmallmatrix*"
    "xalignat"
    "xalignat*"
    "xxalignat"
]

skip_envs = ()